/*
 *  AudioFeature Framework
 *	Compute some low-level audio descriptors.
 *  
 *  Marcelo Cicconet, Visgraf / IMPA
 *	www.impa.br/~cicconet
 *
 *	This framework is distributed under the MIT License.
 *	See notice at the end of this file.
 *
 */

#import "AFAudioDescriptors.h"


@implementation AFAudioDescriptors

// ----------------------------------------------------------------------------------------------------
#pragma mark Initialization
// ----------------------------------------------------------------------------------------------------

- (id)initWithForwardFFTSetupForWindowSize:(UInt32)theWindowSize
{
	if (self = [super init]) {
		windowSize = theWindowSize;
		halfWindowSize = theWindowSize/2;
		
		windowSizeInBytes = windowSize*sizeof(float);
		halfWindowSizeInBytes = halfWindowSize*sizeof(float);
		bufferOfWindowSize = (float *)malloc(windowSizeInBytes);
		bufferOfHalfWindowSize = (float *)malloc(halfWindowSizeInBytes);
		
		length = log2(windowSize);
		fftRadix = kFFTRadix2;
		fftSetup = vDSP_create_fftsetup(length, fftRadix);
		fftDirectionFw = kFFTDirection_Forward;
		dspSplitComplexIO.imagp = (float *)calloc(windowSize,sizeof(float)); // zero initialized vector
		hannWindow = (float *)malloc(windowSize*sizeof(float));
		vDSP_hann_window(hannWindow, windowSize, vDSP_HANN_NORM);
		
		// pre-chroma vector setup
		frequencyLag = 22050.0/(float)halfWindowSize;
		halfNumIndBellowWindow = floorf(32.703194*halfWindowSize/22050.0)+1;
		numIndBellowWindow = 2*halfNumIndBellowWindow;
		hannWindowSize = (numIndBellowWindow-1)*frequencyLag;
		
		// chroma vector setup
//		frequencies = (float *)malloc(120*sizeof(float));
//		frequenciesIntervals = (float *)malloc(121*sizeof(float));
//		binsPerFrequencyInterval = (int *)malloc(120*sizeof(int));
//		for (int i = -45; i < 75; i++) { // 120 piano frequency values
//			frequencies[i+45] = 440.0*powf(2.0, (float)i/12.0); // equivalent midi note: 69+i
//			frequenciesIntervals[i+45] = 440.0*powf(2.0, (float)(i-0.5)/12.0);
//		}
//		int i = 75;
//		frequenciesIntervals[i+45] = 440.0*powf(2.0, ((float)i-0.5)/12.0);
		int i = -45;
		minFrequency = 440.0*powf(2.0, (float)(i-0.5)/12.0);
		i = 75;
		maxFrequency = 440.0*powf(2.0, ((float)i-0.5)/12.0);
		
		// bark scale setup
		barkBandEdges = (UInt32 *)malloc(25*sizeof(UInt32));
		barkBandEdges[0] = 0;
		barkBandEdges[1] = 100;
		barkBandEdges[2] = 200;
		barkBandEdges[3] = 300;
		barkBandEdges[4] = 400;
		barkBandEdges[5] = 510;
		barkBandEdges[6] = 630;
		barkBandEdges[7] = 770;
		barkBandEdges[8] = 920;
		barkBandEdges[9] = 1080;
		barkBandEdges[10] = 1270;
		barkBandEdges[11] = 1480;
		barkBandEdges[12] = 1720;
		barkBandEdges[13] = 2000;
		barkBandEdges[14] = 2320;
		barkBandEdges[15] = 2700;
		barkBandEdges[16] = 3150;
		barkBandEdges[17] = 3700;
		barkBandEdges[18] = 4400;
		barkBandEdges[19] = 5300;
		barkBandEdges[20] = 6400;
		barkBandEdges[21] = 7700;
		barkBandEdges[22] = 9500;
		barkBandEdges[23] = 12000;
		barkBandEdges[24] = 15500;
		barkBandEdges[25] = 22050;
	}
	return self;
}

// ----------------------------------------------------------------------------------------------------
#pragma mark 1-Dimensional Audio Features
// ----------------------------------------------------------------------------------------------------

- (float)loudnessOfAudioWindow:(float *)theAudioWindow
{
	float loudnessBin;
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self powerSpectrum:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	vDSP_sve(bufferOfHalfWindowSize, 1, &loudnessBin, halfWindowSize);
	return loudnessBin;
}

- (float)spectralCentroidOfAudioWindow:(float *)theAudioWindow
{
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self fftMagnitudes:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	vDSP_sve(bufferOfHalfWindowSize, 1, &sumFFTMag, halfWindowSize);
	float specCentBin = 0.0;
	if (sumFFTMag > 0.0) {
		for (UInt32 i = 0; i < halfWindowSize; i++) {
			specCentBin += (float)(i+1)*bufferOfHalfWindowSize[i]/sumFFTMag;
		}
	}
	return specCentBin;
}

- (float)spectralSpreadOfAudioWindow:(float *)theAudioWindow
{
	float specCentBin = [self spectralCentroidOfAudioWindow:theAudioWindow];
	float specSprdBin = 0.0;
	float difference;
	if (sumFFTMag > 0.0) {
		for (UInt32 i = 0; i < halfWindowSize; i++) {
			difference = (float)i+1.0-specCentBin;
			difference *= difference;
			specSprdBin += difference*bufferOfHalfWindowSize[i]/sumFFTMag;
		}
	}
	return specSprdBin;
}

- (float)spectralSkewnessOfAudioWindow:(float *)theAudioWindow
{
	float specCentBin = [self spectralCentroidOfAudioWindow:theAudioWindow];
	float specSprdBin = 0.0;
	float difference;
	if (sumFFTMag > 0.0) {
		for (UInt32 i = 0; i < halfWindowSize; i++) {
			difference = (float)i+1.0-specCentBin;
			difference *= difference;
			specSprdBin += difference*bufferOfHalfWindowSize[i]/sumFFTMag;
		}
	}
	float sigma = sqrtf(specSprdBin);
	float specSknsBin = 0.0;
	if (sumFFTMag > 0.0) {
		for (UInt32 i = 0; i < halfWindowSize; i++) {
			difference = (float)i+1.0-specCentBin;
			difference = difference*difference*difference;
			specSknsBin += difference*bufferOfHalfWindowSize[i]/sumFFTMag;
		}
		specSknsBin = specSknsBin/(sigma*sigma*sigma);
	}
	return specSknsBin;
}

- (float)spectralRollOffOfAudioWindow:(float *)theAudioWindow
{
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self powerSpectrum:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	float loudness;
	vDSP_sve(bufferOfHalfWindowSize, 1, &loudness, halfWindowSize);
	float spectralRollOff = 0.0;
	if (loudness > 0) {
		float rollOff = 0.95*loudness;
		float accumulator = 0.0;
		UInt32 bin = 0;
		while (accumulator < rollOff) {
			accumulator += bufferOfHalfWindowSize[bin];
			bin += 1;
		}
		spectralRollOff = (float)bin;
	}
	return spectralRollOff;
}

- (float)spectralFlatnessOfAudioWindow:(float *)theAudioWindow
{
	float sum, prod, specFlatBin;
	float exp = 1.0/(float)halfWindowSize;
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self fftMagnitudes:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	vDSP_sve(bufferOfHalfWindowSize, 1, &sum, halfWindowSize);
	specFlatBin = 0.0;
	if (sum > 0.0) {
		prod = 1.0;
		for (UInt32 i = 0; i < halfWindowSize; i++) {
			prod *= powf(bufferOfHalfWindowSize[i], exp);
		}
		specFlatBin = halfWindowSize*prod/sum;
	}
	return specFlatBin;
}

- (float)spectralCrestOfAudioWindow:(float *)theAudioWindow
{
	float sum, max, specCrestBin;
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self fftMagnitudes:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	vDSP_sve(bufferOfHalfWindowSize, 1, &sum, halfWindowSize);
	vDSP_maxv(bufferOfHalfWindowSize, 1, &max, halfWindowSize);
	specCrestBin = 0.0;
	if (sum > 0.0) {
		specCrestBin = halfWindowSize*max/sum;
	}
	return specCrestBin;
}

// ----------------------------------------------------------------------------------------------------
#pragma mark N-Dimensional Audio Features
// ----------------------------------------------------------------------------------------------------

// 84-dimensional
- (void)preChroma:(float *)thePreChroma ofAudioWindow:(float *)theAudioWindow
{
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self powerSpectrum:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	[self preChromaVector:thePreChroma fromPowerSpectrum:bufferOfHalfWindowSize];
}

// 12-dimensional
- (void)chroma:(float *)theChroma ofAudioWindow:(float *)theAudioWindow
{
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self powerSpectrum:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	[self chromaVector:theChroma fromPowerSpectrum:bufferOfHalfWindowSize];
}

// 25-dimensional
- (void)barkBands:(float *)theBarkBands ofAudioWindow:(float *)theAudioWindow
{
	memcpy(bufferOfWindowSize, theAudioWindow, windowSizeInBytes);
	[self inPlaceWindowingOfData:bufferOfWindowSize];
	[self powerSpectrum:bufferOfHalfWindowSize ofRealVector:bufferOfWindowSize];
	[self barkScaleWeights:theBarkBands fromPowerSpectrum:bufferOfHalfWindowSize];
}

// ----------------------------------------------------------------------------------------------------
#pragma mark Auxiliary Routines
// ----------------------------------------------------------------------------------------------------

- (void)inPlaceForwardFFTOfComplexVector:(DSPSplitComplex)theComplexVector
{
	vDSP_fft_zip(fftSetup, &theComplexVector, 1, length, fftDirectionFw);
}

- (void)fftMagnitudes:(float *)theMagnitudes ofComplexVector:(DSPSplitComplex)theComplexVector
{
	[self inPlaceForwardFFTOfComplexVector:theComplexVector];
	vDSP_zvabs(&theComplexVector, 1, theMagnitudes, 1, halfWindowSize);
}

- (void)fftMagnitudes:(float *)theMagnitudes ofRealVector:(float *)theRealVector
{
	[self clearImagPartOfSplitComplex];
	dspSplitComplexIO.realp = theRealVector;
	[self fftMagnitudes:theMagnitudes ofComplexVector:dspSplitComplexIO];
}

- (void)powerSpectrum:(float *)thePowerSpectrum ofComplexVector:(DSPSplitComplex)theComplexVector
{
	[self inPlaceForwardFFTOfComplexVector:theComplexVector];
	vDSP_zvmags(&theComplexVector, 1, thePowerSpectrum, 1, halfWindowSize);
}

- (void)powerSpectrum:(float *)thePowerSpectrum ofRealVector:(float *)theRealVector
{
	[self clearImagPartOfSplitComplex];
	dspSplitComplexIO.realp = theRealVector;
	[self powerSpectrum:thePowerSpectrum ofComplexVector:dspSplitComplexIO];
}

- (void)preChromaVector:(float *)thePreChromaVector fromPowerSpectrum:(float *)thePowerSpectrum
{
	// WARNING: halfWindowSize possible values: 512, 1024, 2048, ...
	
	// compute the 84 piano frequencies probability vector (a.k.a. pre-chroma vector)
	float frequency, accumulator, hannValue, currentFrequency;
	float sevenOctavesDensity[84]; // 7*12 = 84
	float max = 0.0;
	UInt32 currentIndex, toTheLeftOfFreqIndex;
	for (SInt32 i = -45; i < 39; i++) { // 84 piano frequency values
		frequency = 440.0*powf(2.0, i/12.0); // equivalent midi note: 69+i
		toTheLeftOfFreqIndex = floorf(frequency/frequencyLag);
		accumulator = 0.0;
		for (UInt8 j = 0; j < numIndBellowWindow; j++) {
			currentIndex = toTheLeftOfFreqIndex-halfNumIndBellowWindow+1+j;
			currentFrequency = currentIndex*frequencyLag;
			hannValue = [self evaluateHannWindowWithCenter:frequency
												   andSize:hannWindowSize
												   atPoint:currentFrequency];
			accumulator = accumulator+hannValue*thePowerSpectrum[currentIndex];
		}
		sevenOctavesDensity[i+45] = accumulator;
		if (accumulator > max) max = accumulator;
	}
	
	// normalize pre chroma vector
	if (max > 0.0) {
		for (UInt8 i = 0; i < 84; i++) {
			thePreChromaVector[i] = sevenOctavesDensity[i]/max;
		}
	}
	else {
		for (UInt8 i = 0; i < 84; i++) {
			thePreChromaVector[i] = 0.0;
		}
	}
}

- (void)chromaVector:(float *)theChromaVector fromPowerSpectrum:(float *)thePowerSpectrum
{
	float frequency, nearestExactFrequency, distance;
	float fIndex;
	int index;
	int binsPerFrequencyInterval[120]; for (int i = 0; i < 120; i++) { binsPerFrequencyInterval[i] = 0; }
	float preChromaVector[120]; for (int i = 0; i < 120; i++) { preChromaVector[i] = 0.0; }
	for (int i = 1; i < halfWindowSize; i++) {
		frequency = 22050.0*(float)i/(float)halfWindowSize;
		if (frequency >= minFrequency && frequency < maxFrequency) {
			fIndex = log2f(powf(frequency/440.0, 12.0))+0.5;
			index = (int)floorf(fIndex);
			nearestExactFrequency = 440.0*powf(2.0, (float)index/12.0);
			distance = fabsf(frequency-nearestExactFrequency);
			binsPerFrequencyInterval[index+45] += 1;
			preChromaVector[index+45] += [self chromaWeightForDistance:distance]*thePowerSpectrum[i];
		}
	}
	for (int i = 0; i < 120; i++) {
		if (binsPerFrequencyInterval[i] > 0) {
			preChromaVector[i] /= (float)binsPerFrequencyInterval[i];
		} else {
			preChromaVector[i] = 0.0;
		}
	}
	float max = 0.0;
	for (int key = 0; key < 12; key++) {
		theChromaVector[key] = 0.0;
		for (int scale = 0; scale < 10; scale++) {
			theChromaVector[key] += preChromaVector[scale*12+key];
		}
		if (theChromaVector[key] > max) max = theChromaVector[key];
	}
	if (max > 0) {
		for (int key = 0; key < 12; key++) {
			theChromaVector[key] /= max;
		}
	}
}

- (void)barkScaleWeights:(float *)theBarkScaleWeights fromPowerSpectrum:(float *)thePowerSpectrum
{
	float accumulator;
	float max = -INFINITY;
	for (UInt32 i = 0; i < 25; i++) {
		UInt32 li = floorf((float)halfWindowSize*(float)barkBandEdges[i]/22050.0);
		UInt32 ui = floorf((float)halfWindowSize*(float)barkBandEdges[i+1]/22050.0);
		accumulator = 0.0;
		for (UInt32 j = li; j < ui; j++) {
			accumulator += thePowerSpectrum[j];
		}
		theBarkScaleWeights[i] = accumulator;
		if (accumulator > max) max = accumulator;
	}
	if (max > 0) {
		for (UInt32 i = 0; i < 25; i++) {
			theBarkScaleWeights[i] /= max;
		}
	}
}

- (float)loudnessFromPowerSpectrum:(float *)thePowerSpectrum
{
	float loudnessBin;
	vDSP_sve(thePowerSpectrum, 1, &loudnessBin, halfWindowSize);
	return loudnessBin;
}

- (float)highFrequencyContentFromPowerSpectrum:(float *)thePowerSpectrum
{
	float hfcBin = 0.0;
	for (UInt32 i = 0; i < halfWindowSize; i++) {
		hfcBin += (thePowerSpectrum[i]*(float)i*(float)i);
	}
	return hfcBin;
}

- (float)chromaWeightForDistance:(float)theDistance
{
	return expf(-theDistance*theDistance/1000.0);
}

- (void)clearImagPartOfSplitComplex
{
	vDSP_vclr(dspSplitComplexIO.imagp, 1, windowSize);
}

- (void)inPlaceWindowingOfData:(float *)theData
{
	vDSP_vmul(theData, 1, hannWindow, 1, theData, 1, windowSize);
}

- (float)evaluateHannWindowWithCenter:(float)theCenter andSize:(float)theSize atPoint:(float)thePoint
{
	return (1.0-cosf(2.0*M_PI*(thePoint-theCenter+theSize/2.0)/theSize))/theSize;
}

- (float)hertzToMel:(float)theHertzFrequency
{
	return 1127.01048*logf(theHertzFrequency/700.0+1.0);
}

// ----------------------------------------------------------------------------------------------------
#pragma mark Cleanup
// ----------------------------------------------------------------------------------------------------

- (void)dealloc
{
	free(barkBandEdges);
	free(bufferOfWindowSize);
	free(bufferOfHalfWindowSize);
	vDSP_destroy_fftsetup(fftSetup);
	free(dspSplitComplexIO.imagp);
	free(hannWindow);
    [super dealloc];
}

@end

/*
 *	Copyright (C) 2011 by Marcelo Cicconet
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a copy
 *	of this software and associated documentation files (the "Software"), to deal
 *	in the Software without restriction, including without limitation the rights
 *	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *	copies of the Software, and to permit persons to whom the Software is
 *	furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *	THE SOFTWARE.
 */