/*
 *  AudioFeature Framework
 *	Compute some low-level audio descriptors.
 *  
 *  Marcelo Cicconet, Visgraf / IMPA
 *	www.impa.br/~cicconet
 *
 *	This framework is distributed under the MIT License.
 *	See notice at the end of this file.
 *
 */

#import "AFAudioPCM.h"


@implementation AFAudioPCM

@synthesize audioFrames, audioLength;

- (id)initWithSoundFromPath:(NSString *)aPath channelNumber:(UInt8)aChannel;
{
	if ((self = [super init])) {
		/* class-specific initialization goes here */
		[self specificInitializationWithSoundFromPath:aPath channelNumber:aChannel];
	}
	return self;
}

- (void)specificInitializationWithSoundFromPath:(NSString *)aPath channelNumber:(UInt8)aChannel
{
	// open audio file for computation purposes
	ExtAudioFileRef inFile;
	
	// obtain filesystem reference to the file using the file path
	CFStringRef filename = CFStringCreateWithCString (NULL, [aPath UTF8String], kCFStringEncodingUTF8);
	CFURLRef inURL = CFURLCreateWithFileSystemPath (NULL, filename, kCFURLPOSIXPathStyle, true);
	
	// first open the input file
	err = ExtAudioFileOpenURL(inURL, &inFile);
	if (err) { printf ("ExtAudioFileOpenURL: %4.4s, %ld\n", (char*)&err, (long int)err); }
	
	UInt64 inputFrames;
	UInt32 size = sizeof(inputFrames);
	err = ExtAudioFileGetProperty(inFile, kExtAudioFileProperty_FileLengthFrames, &size, &inputFrames);
	if (err) { printf ("ExtAudioFileGetProperty FileLengthFrames: %4.4s, %ld\n", (char*)&err, (long int)err); }
	(inputFrames == 0 ?	printf("Number of Frames: Unknown\n") :	printf("Number of Frames: %llu\n", inputFrames));
	
	// get the input file format
	CAStreamBasicDescription inputFormat;
	size = sizeof(inputFormat);
	err = ExtAudioFileGetProperty(inFile, kExtAudioFileProperty_FileDataFormat, &size, &inputFormat);
	if (err) { printf ("ExtAudioFileGetProperty FileDataFormat: %4.4s, %ld\n", (char*)&err, (long int)err); }
	printf("Source File Format: "); inputFormat.Print();
	
	// get and set the client format - it should be lpcm
	CAStreamBasicDescription clientFormat;
	AudioStreamBasicDescription description;
	description.mSampleRate = 44100.0;
	description.mFormatID = kAudioFormatLinearPCM;
	description.mFormatFlags = kLinearPCMFormatFlagIsFloat;
	description.mBytesPerPacket = 8;
	description.mFramesPerPacket = 1;
	description.mBytesPerFrame = 8;
	description.mChannelsPerFrame = 2;
	description.mBitsPerChannel = 32;
	description.mReserved = 0;
	clientFormat.SetFrom(description);
	// clientFormat.SetCanonical(...) leaves the sample rate untouched; we need 44100
	printf("Client Format: "); clientFormat.Print();
	size = sizeof(clientFormat);
	err = ExtAudioFileSetProperty(inFile, kExtAudioFileProperty_ClientDataFormat, size, &clientFormat);
	if (err) { printf ("ExtAudioFileSetProperty ClientDataFormat: %4.4s, %ld\n", (char*)&err, (long int)err); }
	
	// set up buffers
	// for reading purposes...
	UInt32 kSrcBufSize = (UInt32) (0.1*clientFormat.mSampleRate*clientFormat.mChannelsPerFrame*sizeof(float)); // 0.1s of audio data
	float * srcBuffer = (float *) malloc(kSrcBufSize);
	// ...and for computation purposes
	float * audioData = (float *) malloc(inputFrames*clientFormat.mBytesPerFrame);
	
	// do the read operation
	AudioBufferList fillBufList;
	fillBufList.mNumberBuffers = 1;
	fillBufList.mBuffers[0].mNumberChannels = inputFormat.mChannelsPerFrame;
	fillBufList.mBuffers[0].mDataByteSize = kSrcBufSize;
	fillBufList.mBuffers[0].mData = srcBuffer;
	UInt32 iUpperBound, framesPerBuffer = kSrcBufSize/clientFormat.mBytesPerFrame; // number of frames per buffer
	UInt32 i, audioDataCurrentPosition = 0;
	
	printf("\nStart reading audio data\n");
	clock_t start = clock(); // #include <stdio.h> and #include <time.h> are not necessary...
	while (1) {
		err = ExtAudioFileRead(inFile, &framesPerBuffer, &fillBufList);
		if (err) { printf ("ExtAudioFileRead: %4.4s, %ld\n", (char*)&err, (long int)err); }
		iUpperBound = (UInt32) framesPerBuffer*clientFormat.mBytesPerFrame/sizeof(float);
		for (i = 0; i < iUpperBound; i++) {
			audioData[audioDataCurrentPosition+i] = srcBuffer[i];
		}
		audioDataCurrentPosition = audioDataCurrentPosition+iUpperBound;
		// on output, framesPerBuffer = 0 at end-of-file, so the loop must stop
		if (framesPerBuffer == 0) {
			printf("Finished reading. Audio data entries: %lu\n", (long int)audioDataCurrentPosition);
			break;
		}
	}
	printf("Time elapsed: %f sec\n\n", ((double)clock() - start)/CLOCKS_PER_SEC);
	
	// de-interleaving audio and summing across channels to obtain mono audio data
	audioLength = (UInt64) audioDataCurrentPosition/2;
	printf("audioLength = %llu\n", audioLength);
	UInt64 monoAudioDataSize = (UInt64) audioLength*clientFormat.mBytesPerFrame;
	audioFrames = (float *) malloc(monoAudioDataSize);
	
	//	float max = -INFINITY;
	//	float min = INFINITY;
	switch (aChannel) {
		case 0: // channel1 + channel2
			for (i = 0; i < audioLength; i++) {
				audioFrames[i] = audioData[2*i]+audioData[2*i+1];
				//				if (audioData[2*i] > max) max = audioData[2*i];
				//				if (audioData[2*i] < min) min = audioData[2*i];
			}
			break;
		case 1:
			for (i = 0; i < audioLength; i++) {
				audioFrames[i] = audioData[2*i];
			}			
			break;
		case 2:
			for (i = 0; i < audioLength; i++) {
				audioFrames[i] = audioData[2*i+1];
			}
			break;
		default:
			break;
	}
	//	printf("\nmax: %f, min: %f\n\n", max, min);
	
	// close file
	ExtAudioFileDispose(inFile);
	
	// cleanup
	free(srcBuffer);
	free(audioData);
	CFRelease(filename);
	CFRelease(inURL);
}

- (void)dealloc {
	free(audioFrames);
    [super dealloc];
}

@end

/*
 *	Copyright (C) 2011 by Marcelo Cicconet
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a copy
 *	of this software and associated documentation files (the "Software"), to deal
 *	in the Software without restriction, including without limitation the rights
 *	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *	copies of the Software, and to permit persons to whom the Software is
 *	furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *	THE SOFTWARE.
 */