/*
 *  AudioFeature Framework
 *	Compute some low-level audio descriptors.
 *  
 *  Marcelo Cicconet, Visgraf / IMPA
 *	www.impa.br/~cicconet
 *
 *	This framework is distributed under the MIT License.
 *	See notice at the end of this file.
 *
 */

#import "AFFeaturesMatrix.h"


@implementation AFFeaturesMatrix

@synthesize numWindows;

// --------------------------------------------------------------------------------------------------------------
#pragma mark Initialization
// --------------------------------------------------------------------------------------------------------------

- (id)initWithWindowSize:(UInt32)theWindowSize andHopSize:(UInt32)theHopSize forAudioLength:(UInt32)theAudioLength
{
	if ((self = [super init])) {
		windowSize = theWindowSize;
		halfWindowSize = windowSize/2;
		hopSize = theHopSize;
		numWindows = floorf((float)(theAudioLength-theWindowSize)/(float)theHopSize);
		audioDescriptors = [[AFAudioDescriptors alloc] initWithForwardFFTSetupForWindowSize:windowSize];
	}
	return self;
}

// --------------------------------------------------------------------------------------------------------------
#pragma mark 1-Dimensional Features Vector Methods
// --------------------------------------------------------------------------------------------------------------

- (void)spectralFlux:(float *)theSpectralFlux ofAudioSignal:(float *)theAudioSignal
{
	// set up buffers
	float * audioChunk = (float *)malloc(windowSize*sizeof(float));
	float * oneMagSpectrum = (float *)malloc(halfWindowSize*sizeof(float));
	float * anotherMagSpectrum = (float *)malloc(halfWindowSize*sizeof(float));
	float * difMagSpectrum = (float *)malloc(halfWindowSize*sizeof(float));
	
	// compute spectral flux
	UInt8 distType = 0;
	/*
	 *	distType parameter values:
	 *		0: euclidian distance between normalized successive amplitude spectra
	 *		1: 1-norm distance between normalized successive amplitude spectra
	 *		2: one minus normalized cross correlation between two successive amplitude spectra
	 */		
	theSpectralFlux[0] = 0.0; // in fact length(spectral flux) = numWindows-1
	bool currentMagSpectrum = 0; // 0: oneMagSpectrum, 1: anotherMagSpectrum
	float max, sum, prod, binMagSpec, sqrtSum, oneMagSpectrumNorm, anotherMagSpectrumNorm, dotProdMagSpec;
	UInt32 bytesPerWindow = windowSize*sizeof(float);
	for (UInt32 window = 0; window < numWindows; window++) {
		// windowing
		memcpy(audioChunk, &theAudioSignal[window*hopSize], bytesPerWindow);
		[audioDescriptors inPlaceWindowingOfData:audioChunk];
		if (currentMagSpectrum == 0) {
			// fft magnitudes
			[audioDescriptors fftMagnitudes:oneMagSpectrum ofRealVector:audioChunk];
			// max of fft magnitudes
			vDSP_maxv(oneMagSpectrum, 1, &max, halfWindowSize);
			// normalization, if distType is 0 or 1
			if (distType == 0 || distType == 1) {
				for (UInt32 j = 0; j < halfWindowSize; j++) { if (max > 0.0) oneMagSpectrum[j] /= max; }
			} else { // norm of current mag. spectrum, if distType is 2
				oneMagSpectrumNorm = 0.0;
				for (UInt32 j = 0; j < halfWindowSize; j++) {
					binMagSpec = oneMagSpectrum[j];
					oneMagSpectrumNorm += binMagSpec*binMagSpec;
				}
				oneMagSpectrumNorm = sqrtf(oneMagSpectrumNorm);
			}
			// fill spectral flux bin
			if (window > 0) {
				switch (distType) {
					case 0:
						// actual mag. spectrum minus previous mag. spectrum
						vDSP_vsub(anotherMagSpectrum, 1, oneMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// squaring previous result
						vDSP_vsq(difMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// summing entries of previous result
						vDSP_sve(difMagSpectrum, 1, &sum, halfWindowSize);
						// filling current spectral flux bin
						sqrtSum = sqrtf(sum);
						theSpectralFlux[window] = sqrtSum;		
						break;
					case 1:
						// actual mag. spectrum minus previous mag. spectrum
						vDSP_vsub(anotherMagSpectrum, 1, oneMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// compute magnitude of previous result
						vDSP_vabs(difMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// summing entries of previous result
						vDSP_sve(difMagSpectrum, 1, &sum, halfWindowSize);
						// filling current spectral flux bin
						theSpectralFlux[window] = sum;
						break;
					case 2:
						// dot product between actual and previous mag. spectrum
						vDSP_dotpr(oneMagSpectrum, 1, anotherMagSpectrum, 1, &dotProdMagSpec, halfWindowSize);
						// filling current spectral flux bin
						prod = oneMagSpectrumNorm*anotherMagSpectrumNorm;
						if (prod > 0.0) {
							theSpectralFlux[window] = 1.0-dotProdMagSpec/(prod);
						} else {
							theSpectralFlux[window] = 0.0;
						}
						
						break;
					default:
						break;
				}
			}
			currentMagSpectrum = 1;
		} else {
			// fft magnitudes
			[audioDescriptors fftMagnitudes:anotherMagSpectrum ofRealVector:audioChunk];
			// max of fft magnitudes
			vDSP_maxv(anotherMagSpectrum, 1, &max, halfWindowSize);
			// normalization, if distType is 0 or 1
			if (distType == 0 || distType == 1) {
				for (UInt32 j = 0; j < halfWindowSize; j++) { if (max > 0.0) anotherMagSpectrum[j] /= max; }
			} else { // norm of current mag. spectrum, if distType is 2
				anotherMagSpectrumNorm = 0.0;
				for (UInt32 j = 0; j < halfWindowSize; j++) {
					binMagSpec = anotherMagSpectrum[j];
					anotherMagSpectrumNorm += binMagSpec*binMagSpec;
				}
				anotherMagSpectrumNorm = sqrtf(anotherMagSpectrumNorm);
			}
			// fill spectral flux bin
			if (window > 0) {
				switch (distType) {
					case 0:
						// actual mag. spectrum minus previous mag. spectrum
						vDSP_vsub(oneMagSpectrum, 1, anotherMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// squaring previous result
						vDSP_vsq(difMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// summing entries of previous result
						vDSP_sve(difMagSpectrum, 1, &sum, halfWindowSize);
						// filling current spectral flux bin
						sqrtSum = sqrtf(sum);
						theSpectralFlux[window] = sqrtSum;
						break;
					case 1:
						// actual mag. spectrum minus previous mag. spectrum
						vDSP_vsub(oneMagSpectrum, 1, anotherMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// compute magnitude of previous result
						vDSP_vabs(difMagSpectrum, 1, difMagSpectrum, 1, halfWindowSize);
						// summing entries of previous result
						vDSP_sve(difMagSpectrum, 1, &sum, halfWindowSize);
						// filling current spectral flux bin
						theSpectralFlux[window] = sum;
						break;
					case 2:
						// dot product between actual and previous mag. spectrum
						vDSP_dotpr(anotherMagSpectrum, 1, oneMagSpectrum, 1, &dotProdMagSpec, halfWindowSize);
						// filling current spectral flux bin
						prod = anotherMagSpectrumNorm*oneMagSpectrumNorm;
						if (prod > 0.0) {
							theSpectralFlux[window] = 1.0-dotProdMagSpec/(prod);
						} else {
							theSpectralFlux[window] = 0.0;
						}
						break;
					default:
						break;
				}
			}
			currentMagSpectrum = 0;
		}
	}
	
	// cleanup
	free(audioChunk);
	free(oneMagSpectrum);
	free(anotherMagSpectrum);
	free(difMagSpectrum);
}

- (void)loudness:(float *)theLoudness ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theLoudness[window] = [audioDescriptors loudnessOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralCentroid:(float *)theSpectralCentroid ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralCentroid[window] = [audioDescriptors spectralCentroidOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralSpread:(float *)theSpectralSpread ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralSpread[window] = [audioDescriptors spectralSpreadOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralSkewness:(float *)theSpectralSkewness ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralSkewness[window] = [audioDescriptors spectralSkewnessOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralRollOff:(float *)theSpectralRollOff ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralRollOff[window] = [audioDescriptors spectralRollOffOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralFlatness:(float *)theSpectralFlatness ofAudioSignal:(float *)theAudioSignal
{	
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralFlatness[window] = [audioDescriptors spectralFlatnessOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

- (void)spectralCrest:(float *)theSpectralCrest ofAudioSignal:(float *)theAudioSignal
{
	for (UInt32 window = 0; window < numWindows; window++) {
		theSpectralCrest[window] = [audioDescriptors spectralCrestOfAudioWindow:&theAudioSignal[window*hopSize]];
	}
}

// --------------------------------------------------------------------------------------------------------------
#pragma mark N-Dimensional Features Vector Methods
// --------------------------------------------------------------------------------------------------------------

- (void)preChromagram:(AFFloatsMatrix *)thePreChromagram ofAudioSignal:(float *)theAudioSignal
{
	float preChromaVector[84];
	for (UInt32 window = 0; window < numWindows; window++) {
		[audioDescriptors preChroma:preChromaVector ofAudioWindow:&theAudioSignal[window*hopSize]];
		for (UInt8 i = 0; i < 84; i++) {
			[thePreChromagram setValue:preChromaVector[i] forRow:83-i andCol:window];
		}
	}
}

- (void)chromagram:(AFFloatsMatrix *)theChromagram ofAudioSignal:(float *)theAudioSignal
{
	float chromaVector[12];
	for (UInt32 window = 0; window < numWindows; window++) {
		[audioDescriptors chroma:chromaVector ofAudioWindow:&theAudioSignal[window*hopSize]];
		for (UInt8 i = 0; i < 12; i++) {
			[theChromagram setValue:chromaVector[i] forRow:11-i andCol:window];
		}
	}
}

- (void)barkBands:(AFFloatsMatrix *)theBarkBands ofAudioSignal:(float *)theAudioSignal
{
	float barkBands[25];
	for (UInt32 window = 0; window < numWindows; window++) {
		[audioDescriptors barkBands:barkBands ofAudioWindow:&theAudioSignal[window*hopSize]];
		for (UInt32 band = 0; band < 25; band++) {
			[theBarkBands setValue:barkBands[band] forRow:24-band andCol:window];
		}
	}
}

- (void)chromaAndLoudness:(AFFloatsMatrix *)theChromaAndLoudness ofAudioSignal:(float *)theAudioSignal
{
	// setup buffers
	float * audioChunk = (float *)malloc(windowSize*sizeof(float));
	float * powerSpectrum = (float *)malloc(halfWindowSize*sizeof(float));
	float chromaVector[12];
	float * loudnessBuffer = (float *)malloc(numWindows*sizeof(float));
	
	// compute chroma matrix with loudness
	float sum;
	float maxLoud = 0.0;
	UInt32 bytesPerWindow = windowSize*sizeof(float);
	for (UInt32 window = 0; window < numWindows; window++) {
		// windowing
		memcpy(audioChunk, &theAudioSignal[window*hopSize], bytesPerWindow);
		[audioDescriptors inPlaceWindowingOfData:audioChunk];
		// power spectrum
		[audioDescriptors powerSpectrum:powerSpectrum ofRealVector:audioChunk];
		// loudness
		vDSP_sve(powerSpectrum, 1, &sum, halfWindowSize);
		loudnessBuffer[window] = sum;
		if (sum > maxLoud) maxLoud = sum;
		// chroma vector
		[audioDescriptors chromaVector:chromaVector fromPowerSpectrum:powerSpectrum];
		for (UInt8 i = 0; i < 12; i++) {
			[theChromaAndLoudness setValue:chromaVector[i] forRow:11-i andCol:window];
		}
	}
	if (maxLoud > 0.0) {
		float value;
		for (UInt32 window = 0; window < numWindows; window++) {
			for (UInt8 i = 0; i < 12; i++) {
				value = [theChromaAndLoudness valueAtRow:i andCol:window];
				[theChromaAndLoudness setValue:loudnessBuffer[window]/maxLoud*value forRow:i andCol:window];
			}
		}
	}
	
	// cleanup
	free(loudnessBuffer);
	free(audioChunk);
	free(powerSpectrum);
}

// --------------------------------------------------------------------------------------------------------------
#pragma mark Cleanup
// --------------------------------------------------------------------------------------------------------------

- (void)dealloc
{
	[audioDescriptors release];
	[super dealloc];
}

@end

/*
 *	Copyright (C) 2011 by Marcelo Cicconet
 *
 *	Permission is hereby granted, free of charge, to any person obtaining a copy
 *	of this software and associated documentation files (the "Software"), to deal
 *	in the Software without restriction, including without limitation the rights
 *	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *	copies of the Software, and to permit persons to whom the Software is
 *	furnished to do so, subject to the following conditions:
 *
 *	The above copyright notice and this permission notice shall be included in
 *	all copies or substantial portions of the Software.
 *
 *	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *	THE SOFTWARE.
 */